/******************************************************************************/
/*                                                                            */
/* Licensed Materials - Property of IBM                                       */
/* (C) Copyright IBM Corp. 2000                                               */
/*                                                                            */
/* Program name    : alarm.c                                                  */
/* Output name     : none                                                     */
/*                                                                            */
/* Compile options : /VERBOSE /LD /MT                                         */
/*                   INCLUDE must contain <install dir>\inlude\plugin         */
/* Link options    : /DLL                                                     */
/*                   LIB must contain <install dir>\bin\imbdfplg.lib          */
/*                                                                            */
/* Description     : These utility functions are used within the treeWalk     */
/*                   plug-in node.                                            */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/* Version 1.0                                                                */
/* Author: Mike Brady                    Date: Feb 2000                       */
/*                                                                            */
/******************************************************************************/

/* Include Headers List */
#include "aggregator.h" /* Specific header for this dll                            */
#include <malloc.h>
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "threads.h"
#include <time.h>
#include <cmqc.h>

void genGMTDateAndTime(OUT char *dateString,
                       OUT char *timeString);

/*
 * Reset the alarm element for a new or modified aggregate 
 */
void setAggregateReplyAlarm(IN aggregateInstance_t *aggregateP,
	  	                    IN aggregateMember_t *memberP,
							IN bool_t recoveryProcessing)
{

	/*
	 * Assumption:
	 *
	 * On entry to this function, the caller has got the aggregate chain locked
	 */
	time_t currentTime;
	int timeoutInterval;
	aggregateInstance_t *currentAggregateP;

    TRACE_ENTRY("setAggregateReplyAlarm", ("aggregateP=%.8x, memberP=%.8x, "
		                                   "recoveryProcessing=%s",
		                                    aggregateP, memberP,
											recoveryProcessing ? "TRUE" : "FALSE"));

	/* if aggregate is already part of the alarm chain, remove it */
	if (aggregateP->timeoutTime)
	{
		TRACE_MID(TRACE_INFO, ("Aggregate alarm pointers: prev=%.8x, next=%.8x",
			                   aggregateP->prevAlarmP,
							   aggregateP->nextAlarmP));
		if (!recoveryProcessing || 
		    recoveryProcessing && 
			(aggregateP->prevAlarmP ||
			 aggregateP->nextAlarmP))
		{
			/* remove from alarm chain */
			TRACE_MID(TRACE_INFO, ("Removing aggregate descriptor %.8x from alarm chain", 
								   aggregateP));
			REMOVE_FROM_ALARM_CHAIN(aggregateP);
		}
	}

	if (!recoveryProcessing)
	{
		/* calculate the new reply timeout interval */
		timeoutInterval = aggregateP->deployedAggregateP->timeoutInterval;
		TRACE_MID(TRACE_INFO, ("Timeout interval is %d", timeoutInterval));
		time(&currentTime);
		aggregateP->timeoutTime = currentTime + timeoutInterval;
		TRACE_MID(TRACE_INFO, ("Timeout value is %d", aggregateP->timeoutTime));
		memberP->requestArrivalTime = currentTime;
	}

	/* find position in alarm chain for this aggregate based on its expiry time */
    currentAggregateP = &alarmAnchor;
    TRACE_MID(TRACE_INFO, ("Starting search of alarm chain at anchor (%.8x)",
		                   currentAggregateP));

	while(TRUE)
	{
        TRACE_MID(TRACE_INFO, ("Checking aggregate descriptor at %.8x "
			                   "(timeout value = %d)", currentAggregateP,
							   currentAggregateP->timeoutTime));

		if (!currentAggregateP->nextAlarmP ||
			currentAggregateP->nextAlarmP->timeoutTime > aggregateP->timeoutTime)
		{
            TRACE_MID(TRACE_INFO, ("Inserting alarm after aggregate at %.8x",
				                   currentAggregateP));
			break;
		}

		currentAggregateP = currentAggregateP->nextAlarmP;
	}

    /* add after current alarm member */
	INSERT_AFTER_IN_ALARM_CHAIN(currentAggregateP, aggregateP);
 	
    TRACE_RETURN_VOID(RSN_001);
}

typedef struct handleEntry {
	char *queueName;
	MQHOBJ queueHandle;
} handleEntry_t;

handleEntry_t *queueHandleTable = NULL;
int numQueueHandleSlots = 0,
	numQueueHandleEntries = 0;

/******************************************************************************/
/* processAlarms                                                              */
/******************************************************************************/
threadFunction_t processAlarms(void *dataP)
{
    MQLONG   compCode,                  /* completion code               */
             reason;                    /* reason code                   */
	MQHCONN qmConn = 0;
	char errorMessage[1000];
	time_t currentTime;
	int timeToWait;
	aggregateInstance_t *thisAlarmP;
	aggregateMember_t *memberP;
	int retcode;

    TRACE_ENTRY("processAlarms", ("dataP=%.8x", dataP));

	/* get handle to queue manager */
    TRACE_MID(TRACE_INFO, ("About to issue MQCONN for '%s'",
		                    queueManagerName));
	MQCONN(queueManagerName,
		   &qmConn,                 /* connection handle */
		   &compCode,               /* completion code */
		   &reason);                /* reason code */

	if (compCode == MQCC_FAILED)
	{
		/* create error message to pass back to parent */
		sprintf(errorMessage, "MQCONN for queue manager '%s' "
                           "failed with reason code %ld",
                           queueManagerName,
                           reason);

        TRACE_MID(TRACE_INFO, ("%s", errorMessage));
        alarmThreadErrorMessage = strdup(errorMessage);

		/* lock the alarm thread sync mutex and post the parent thread */
	    threadMutexLock(alarmThreadSyncMutex);
		threadCondSignal(alarmThreadSyncCondVar);
	    threadMutexUnlock(alarmThreadSyncMutex);

        TRACE_RETURN_TYPE(RSN_002, FUNCTION_RETURN_TYPE, 0);
	}

	/* let the parent thread know that we're up and running */
	alarmThreadInitialised = TRUE;

	/* lock the alarm thread sync mutex and post the parent thread */
    TRACE_MID(TRACE_INFO, ("About to signal parent for initialisation complete"));
	threadMutexLock(alarmThreadSyncMutex);
	threadCondSignal(alarmThreadSyncCondVar);
	threadMutexUnlock(alarmThreadSyncMutex);

	/* 
	 * The aggregate chain and the alarm chain are closely related, so we lock
	 * both via a single mutex.
	 */
	TRACE_MID(TRACE_INFO, ("About to lock aggregateChainMutex"));
	threadMutexLock(aggregateChainMutex);

	/* keep processing alarm chain until requested to shutdown */
	while(!shutdownRequested)
	{
        TRACE_MID(TRACE_INFO, ("Processing alarm element chain"));

		thisAlarmP = alarmAnchor.nextAlarmP;

		/* process any expired alarm elements */
	    while(thisAlarmP)
		{
			/* get current time */
			time(&currentTime);
            TRACE_MID(TRACE_INFO, ("Current time is %d", currentTime));

 		    /* run down the list of alarm elements */
            TRACE_MID(TRACE_INFO, ("Processing alarm element at %.8x, timeoutTime=%d",
				                    thisAlarmP, thisAlarmP->timeoutTime));

			/* 
			 * if the ReplyTimeoutInterval has expired, then process any
			 * messages we've received up till this point.
			 */
			if (thisAlarmP->timeoutTime <= currentTime)
			{
				if (traceEnabled)
				{
					char hexBuffer[2*sizeof(thisAlarmP->aggregateId) + 1];
					dispHex((char *)(thisAlarmP->aggregateId),
							MQ_CORREL_ID_LENGTH,
							hexBuffer);
					TRACE_MID(TRACE_INFO, ("Aggregate instance %.8x has expired (%s)",
											thisAlarmP, hexBuffer));
				}

				thisAlarmP->timedOut = TRUE;

				/* process expired alarm element */
				REMOVE_FROM_LINKED_LIST(thisAlarmP);
				REMOVE_FROM_ALARM_CHAIN(thisAlarmP);

				/* for each member, remove its msg index entry */
				memberP = thisAlarmP->memberAnchor.nextP;

				TRACE_MID(TRACE_INFO, ("Removing index entries for members"));
				while(memberP)
				{
					TRACE_MID(TRACE_INFO, ("Processing member at %.8x",
										   memberP));
					if (memberP->indexP)
					{
						TRACE_MID(TRACE_INFO, ("Removing index entry at %.8x",
												memberP->indexP));
						removeMsgIdIndex(memberP->indexP);
						memberP->indexP = NULL;
					}
					memberP = memberP->nextP;
				}

				/* add to the timed out chain */
				INSERT_AFTER_IN_LINKED_LIST(timedOutAnchorP, thisAlarmP);
				TRACE_MID(TRACE_INFO, ("Added to timed out chain"));

				/* release the lock to allow other threads to run */
				TRACE_MID(TRACE_INFO, ("About to unlock aggregateChainMutex"));
				threadMutexUnlock(aggregateChainMutex);

				retcode = notifyAggregateReplyNodeOfExpiredAggregate(qmConn,
					                                                 thisAlarmP);

				if (retcode != TT_SUCCESS)
				{
				    /* not sure what to do if this function fails... */
				}

				/* 
				 * The alarm chain may have been modified by another thread, so 
				 * start processing at head of chain again.
				 */
	            TRACE_MID(TRACE_INFO, ("About to lock aggregateChainMutex"));
	            threadMutexLock(aggregateChainMutex);

				thisAlarmP = alarmAnchor.nextAlarmP;
				continue;
			}

			/* alarms are in order of expiry time...so nothing more to do here */
			break;
		}

		/* reset the alarm thread to pop at next expiry interval */
		if (alarmAnchor.nextAlarmP)
		{
			/* reset the alarm for the next element (milliseconds) */
			timeToWait = 
				(alarmAnchor.nextAlarmP->timeoutTime - currentTime) * 1000;
            TRACE_MID(TRACE_INFO, ("Next expiry interval time is %d", timeToWait));
		}
		else
		{
			timeToWait = INFINITE;
            TRACE_MID(TRACE_INFO, ("Next expiry interval time is INFINITE"));
		}

	    /* wait for next alarm to pop or a new element to appear on the chain */
        TRACE_MID(TRACE_INFO, ("Alarm thread about to sleep for expiry interval"));
		retcode = threadCondTimedWait(alarmThreadWorkCondVar, timeToWait);
        TRACE_MID(TRACE_INFO, ("Alarm thread woken due to %s",
							   retcode == THREAD_TIMED_OUT ? "time out" : "signal"));
	}

	if (qmConn)
	{
		if (numQueueHandleSlots)
		{
			TRACE_MID(TRACE_INFO, ("Freeing up queue handles"));

			/* see if we've already got a handle to this queue */
			numQueueHandleEntries = 0;  /* remove all slots */
			numQueueHandleSlots = 0;  /* remove all slots */
			free(queueHandleTable);
			queueHandleTable = NULL;
		}

		TRACE_MID(TRACE_INFO, ("About to issue MQDISC for '%s'",
								queueManagerName));
		MQDISC(&qmConn,                 /* connection handle */
			   &compCode,               /* completion code */
			   &reason);                /* reason code */
	}

    TRACE_MID(TRACE_INFO, ("Alarm thread terminating"));

	threadMutexUnlock(aggregateChainMutex);

    TRACE_RETURN_TYPE(RSN_003, FUNCTION_RETURN_TYPE, 0);
}

/******************************************************************************/
/* notifyAggregateReplyNodeOfExpiredAggregate                                 */
/******************************************************************************/
int notifyAggregateReplyNodeOfExpiredAggregate(IN MQHCONN qmConn,
											   IN aggregateInstance_t *aggregateP)
{
    MQLONG   compCode,                  /* completion code               */
             reason;                    /* reason code                   */
    MQLONG   openOptions;               /* MQOPEN options                */
	MQHOBJ queueHandle;
    MQOD od = {MQOD_DEFAULT};           /* Object Descriptor */
    MQMD md = {MQMD_DEFAULT};           /* Message Descriptor */
    MQPMO pmo = {MQPMO_DEFAULT};        /* put message options */
	char buffer[1],
		 execGroupKey[10],
		 aggNameKey[10],
		 aggNameMB[MAX_STRING_SIZE+1],
		 queueNameMB[MAX_STRING_SIZE+1],
		 *replyToQueueP,
		 dateString[4+2+2+1],
		 timeString[2+2+2+2+1];
	int i,
		keyLen;
	msgIdIndexElement_t *indexP;

    TRACE_ENTRY("notifyAggregateReplyNodeOfExpiredAggregate", ("qmConn=%.8x, aggregateP=%.8x",
		                                                        qmConn,
																aggregateP));

	/* assume caller has NOT got the aggregate chain locked on entry */


	/* build a timed out message to pass to the AggregateReply node */

	/* send timed out message to the reply node input queue */
	UCsprintf(queueNameMB, "%S", aggregateP->deployedAggregateP->replyToQueueName);
	replyToQueueP = queueNameMB;

	/* see if we've already got a handle to this queue */
	queueHandle = 0;
    TRACE_MID(TRACE_INFO, ("Searching for stored queue handle for '%s'",
		                    replyToQueueP));
	for (i=0; i<numQueueHandleEntries; i++)
	{
        TRACE_MID(TRACE_INFO, ("Processing stored queue handle for '%s'",
			                    (queueHandleTable+i)->queueName));
		if (!strcmp((queueHandleTable+i)->queueName,
			         replyToQueueP))
		{
            TRACE_MID(TRACE_INFO, ("Found stored queue handle"));
			queueHandle = (queueHandleTable+i)->queueHandle;
		}
	}

	if (!queueHandle)
	{
		/* open the queue and store the handle */

		/* initialise the object descriptor with the queue name */
		strncpy(od.ObjectName, 
                replyToQueueP,
				(size_t)MQ_Q_NAME_LENGTH);

		/* set open mode for writing messages */
		openOptions = MQOO_OUTPUT;
		openOptions += MQOO_SET_ALL_CONTEXT;       /* we need to set PutApplName */
		openOptions += MQOO_FAIL_IF_QUIESCING;     /* but not if MQM stopping */
		 
		TRACE_MID(TRACE_INFO, ("About to issue MQOPEN for '%s'",
								replyToQueueP));

		MQOPEN(qmConn,              /* connection handle */
			   &od,                 /* object descriptor for queue */
			   openOptions,         /* open options */
			   &queueHandle,        /* object handle */
			   &compCode,           /* completion code */
			   &reason);            /* reason code */
            
		if (compCode == MQCC_FAILED)
		{
			TRACE_MID(TRACE_ERROR, ("MQOPEN for '%s' failed with "
								  "reason code %ld",
								  replyToQueueP,
								  reason));
            TRACE_RETURN(RSN_004, TT_FAILED);
		}

		/* increase handle table if full */
		if (numQueueHandleSlots == numQueueHandleEntries)
		{
			numQueueHandleSlots +=10;  /* increase by 10 slots */
	        TRACE_MID(TRACE_INFO, ("Increasing handle table to %d slots",
							       numQueueHandleSlots));
            queueHandleTable = (handleEntry_t *)realloc(queueHandleTable,
							   (numQueueHandleSlots)*(sizeof(handleEntry_t)));
		}

		/* add queue handle to table */
        TRACE_MID(TRACE_INFO, ("Queue handle for '%s' added to slot %d",
						       replyToQueueP,
							   numQueueHandleEntries));
		(queueHandleTable+numQueueHandleEntries)->queueName = strdup(replyToQueueP);
		(queueHandleTable+numQueueHandleEntries)->queueHandle = queueHandle;
        numQueueHandleEntries++;
	}

	/* send a timed out message with 0 length data to reply queue */
    md.Persistence = MQPER_PERSISTENT;

	/*
	 * It would be desirable to be able to use the single SYSTEM.BROKER.TIMER.QUEUE for
	 * all plugins requiring a similar timer function.  To allow the MQInput nodes to 
	 * differentiate between an Aggregator timer message and another plugin timer message,
	 * we will use a 4 letter prefix ('Agg-') in the CorrelId and the corresponding MQInput 
	 * node will have its Advanced.CorrelId value set to the hex value 4167672e.
	 *
	 * Furthermore, it would be useful to be able to share the same SYSTEM.BROKER.TIMER.QUEUE
	 * between different deployed aggregates running on the same broker.  For situations
	 * where more than one deployed aggregate exists on a one or more Execution Groups,
	 * a further qualification of the CorrelId is required.  The proposed solution is to 
	 * build the CorrelId from the prefix already discussed (Agg-) together with a 
	 * combination of the Aggregate name and Execution Group values.
	 *
	 * However, since this requires manipulation of the 
	 * AggregateReply.MQInput.Advanced.CorrelId property value, a customiser will be 
	 * required.  This has not yet been written.
	 *
     * In the interim, should you require multiple aggregate configurations to be
	 * deployed simultaneously on one or more Execution Groups on the same broker, you must 
	 * manually set the SpAggregateReply CorrelIdValue property. 
	 */

	/* set correl id to 'Agg-' - to indicate an aggregator timer msg */
	memset(md.CorrelId, 0, sizeof(md.CorrelId));
	memcpy(md.CorrelId, AGG_TIMER_ID_MB, strlen(AGG_TIMER_ID_MB));  /* 0x4167672e */

	/* 
	 * If the SetTimeoutCorrelId property is set to TRUE...
	 * append 10 chars of AggregateName and 10 chars of Execution Group name.
	 * If either value consists of more than 10 chars, then take the first and last 5
	 * characters.
	 */

	if (aggregateP->deployedAggregateP->setTimeoutCorrelId)
	{
		/* build aggregate name key */
		memset(aggNameKey, ' ', sizeof(aggNameKey));
        UCsprintf(aggNameMB, "%S", aggregateP->deployedAggregateP->name);
		keyLen = strlen(aggNameMB);

		if (keyLen > 10)
		{
			/* take first and last 5 chars */
			memcpy(aggNameKey, aggNameMB, 5);
			memcpy(aggNameKey+5, aggNameMB+keyLen-5, 5);
		}
		else
		{
			memcpy(aggNameKey, aggNameMB, keyLen);
		}

		/* build execution group key */
		keyLen = strlen(executionGroupName);
		memset(execGroupKey, ' ', sizeof(execGroupKey));

		if (keyLen > 10)
		{
			/* take first and last 5 chars */
			memcpy(execGroupKey, executionGroupName, 5);
			memcpy(execGroupKey+5, executionGroupName+keyLen-5, 5);
		}
		else
		{
			memcpy(execGroupKey, executionGroupName, keyLen);
		}

		memcpy(md.CorrelId+4, aggNameKey, 10);
		memcpy(md.CorrelId+14, execGroupKey, 10);

		TRACE_MID(TRACE_INFO, ("Setting Correl id to..."));
		TRACE_MIDHEX(TRACE_INFO, md.CorrelId, MQ_CORREL_ID_LENGTH);
	}

	/* set PutApplName value to 'AggregatorTimerThread' */
	memcpy(md.PutApplName, TIMER_THREAD_ID_MB, strlen(TIMER_THREAD_ID_MB));

	/* make sure the PutApplName value is used on input */
	pmo.Options += MQPMO_SET_ALL_CONTEXT;

	/* set the PutDate, PutTime and PutApplType values */
	md.PutApplType = MQAT_DEFAULT;

    genGMTDateAndTime(dateString, timeString);
    TRACE_MID(TRACE_INFO, ("Date string: '%s'", dateString));
    TRACE_MID(TRACE_INFO, ("Time string: '%s'", timeString));
	memcpy(md.PutDate, dateString, sizeof(md.PutDate));
	memcpy(md.PutTime, timeString, sizeof(md.PutTime));
	
    TRACE_MID(TRACE_INFO, ("About to write timed out message"));
    
	/* 
	 * Make sure we lock the aggregate chain to prevent the thread processing the
	 * reply queue from proceeding until we've updated our index entries
	 */
    TRACE_MID(TRACE_INFO, ("About to lock aggregateChainMutex"));
    threadMutexLock(aggregateChainMutex);    

    /* put the message on the queue */
    MQPUT(qmConn,                   /* connection handle */
          queueHandle,              /* object handle */
          &md,                      /* message descriptor */
          &pmo,                     /* put options */
          0,                        /* zero msg length */
          buffer,                   /* dummy message buffer */
          &compCode,                /* completion code */
          &reason);                 /* reason code */
        
	if (compCode != MQCC_OK)
	{
		TRACE_MID(TRACE_ERROR, ("MQPUT for '%s' failed with "
							  "reason code %ld",
							  replyToQueueP,
							  reason));
        TRACE_MID(TRACE_INFO, ("About to unlock aggregateChainMutex"));
        threadMutexUnlock(aggregateChainMutex);    
        TRACE_RETURN(RSN_005, TT_FAILED);
	}
    TRACE_MID(TRACE_INFO, ("Timed out message written to queue"));

	/* store the message id in the msg id index */
	indexP = updateMsgIdIndex((char *)md.MsgId, &msgIdIndexAnchor);
	indexP->memberP = (aggregateMember_t *)aggregateP; 

	/* now unlock the chain to allow reply queue processing thread to proceed */
    TRACE_MID(TRACE_INFO, ("About to unlock aggregateChainMutex"));
    threadMutexUnlock(aggregateChainMutex);    

    TRACE_RETURN(RSN_006, TT_SUCCESS);
}

/*********************/
/* genGMTDateAndTime */
/*********************/
#include <time.h>
void genGMTDateAndTime(OUT char *dateString,
                       OUT char *timeString)
{
    time_t thisTime;
    struct tm *newTime;

	TRACE_ENTRY("genGMTDateAndTime", (" "));

    /* get the date and time info */
    time(&thisTime);
    newTime = gmtime(&thisTime);

    /* 
     * The gmtime function produces a structure containing values for 
     * day, month, year, etc. as UTC
     *
     * NB: will return null pointer for dates prior to midnight 1 Jan 1970
     */

    if (newTime)
    {
        /* extract the bits we need */

        /* year is value less 1900, so we need to qualify it */
        if (newTime->tm_year < 90)
        {
            newTime->tm_year += 2000;
        }
        else
        {
            newTime->tm_year += 1900;
        }

        /* month is 0 - 11, so we bump it by one */
        newTime->tm_mon++;

        sprintf(dateString, "%04.4d%02.2d%02.2d",
                            newTime->tm_year,
                            newTime->tm_mon,
                            newTime->tm_mday);
        sprintf(timeString, "%02.2d%02.2d%02.2d00", /* tenths=0, hundredths=0 */
                            newTime->tm_hour,
                            newTime->tm_min,
                            newTime->tm_sec);
    }
    else
    {
        TRACE_MID(TRACE_ERROR, ("Error extracting time via localtime"));
        dateString = "";
        timeString = "";
    }

    TRACE_RETURN_VOID(RSN_007)
}

void runTheChain()
{
#if 0
	aggregateInstance_t *aggregateP;
	
	if (!deployedAggregateAnchor.nextP)
		return;

    aggregateP = deployedAggregateAnchor.nextP->aggregateAnchor.nextP;

    TRACE_MID(TRACE_INFO, ("*********************************************************"));

    TRACE_MID(TRACE_INFO, ("* Aggregate Chain       *"));
	/* print out contents of aggregate instance chain */
	while(aggregateP)
	{
	    TRACE_MID(TRACE_INFO, ("* aggregate at %.8x, prevP=%.8x, nextP=%.8x *", 
			                   aggregateP, aggregateP->prevP, aggregateP->nextP));
		aggregateP = aggregateP->nextP;
	}
    TRACE_MID(TRACE_INFO, ("*********************************************************"));
#else

	msgIdIndexElement_t *currentIndexP;
	char buffer[50];
	int i;

    TRACE_MID(TRACE_INFO, ("*********************************************************"
		                   "*******************"));

    TRACE_MID(TRACE_INFO, ("* MsgId Index Chain                                      "
		                   "                  *"));

	currentIndexP = &msgIdIndexAnchor;

	i = 0;
	while(currentIndexP)
	{
		memset(buffer, 0, sizeof(buffer));
		dispHex((char *)(currentIndexP->msgId), MQ_MSG_ID_LENGTH, buffer);
	    TRACE_MID(TRACE_INFO, ("* %.3d index at %.8x - %s *", i++, currentIndexP, buffer));
		currentIndexP = currentIndexP->nextP;
	}
    TRACE_MID(TRACE_INFO, ("*********************************************************"
		                   "*******************"));
#endif

	return;
}

