/******************************************************************************/
/*                                                                            */
/* Licensed Materials - Property of IBM                                       */
/* (C) Copyright IBM Corp. 2000                                               */
/*                                                                            */
/* Program name    : destination.c                                            */
/* Output name     : none                                                     */
/*                                                                            */
/* Compile options : /VERBOSE /LD /MT                                         */
/*                   INCLUDE must contain <install dir>\inlude\plugin         */
/* Link options    : /DLL                                                     */
/*                   LIB must contain <install dir>\bin\imbdfplg.lib          */
/*                                                                            */
/* Description     : Contains functions for building destination lists.       */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/* Version 1.0                                                                */
/* Author: Mike Brady                    Date: Jun 2000                       */
/*                                                                            */
/******************************************************************************/

/* Include Headers List */
#include "aggregator.h"      /* Specific header for this dll                       */
#include <malloc.h>
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <BipCni.h>		/* Function prototypes for Plug-in Node functions     */


/******************************************************************************/
/* Node Utility Functions                                                     */
/*                                                                            */
/* evaluateAggregateReply                                                     */
/*                                                                            */
/******************************************************************************/

/******************************************************************************/
/* addAddressToDestinationList                                           */
/******************************************************************************/
int addAddressToDestinationList(IN CciMessage *destinationList,
								IN CciChar *queueName,
								IN CciChar *queueManagerName,
								OUT CciMessage **newDestinationListP)
{
	CciMessage *newDestList;
	int rc;
	CciMessageContext *msgContextP;
	void *msgRootP,
		 *destinationP,
		 *mqDestinationListP,
	     *destinationDataP,
		 *qmnP,
		 *qnP;

    TRACE_ENTRY("addAddressToDestinationList", ("destinationList=%.8x, "
		                                    "queueName=%S, queueManagerName=%S",
											destinationList,
											queueName,
											queueManagerName));

	/* 
	 * The 'Using the Control Center' manual briefly describes the structure of the
	 * destination list.  According to this doco, the list consists of a tree structure
	 * similar to that shown below.
	 *                      root
	 *                        |
	 *                 DestinationList
	 *                        |
	 *                   Destination
	 *                        |
	 *           ----------------------------------
	 *          |             |                    |  
     *       transport_x   MQDestinationList            transport_y
	 *                        |
	 *             ---------------------------
	 *            |                           |
	 *          Defaults               DestinationData
	 *                                   |
	 *                                   |
	 *                                   |- queueManagerName
	 *                                   |- queueName
	 *                                   |
	 *                                   |- etc...
	 */

	/* 
	 * The destination list is held as a message structure.  In keeping with the 
	 * required MQSI message processing model, we must copy the original destination
	 * list before making changes.
	 */

	/* create a new destination list message using the input message as a template */
	msgContextP = cniGetMessageContext(&rc, destinationList);
	if (rc != CCI_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to get message context "
			                    "(cniGetMessageContext rc=%d)", rc));
		TRACE_RETURN(RSN_001, TT_FAILED_TO_CREATE_DESTINATION_LIST);
	}
	TRACE_MID(TRACE_INFO, ("Message context got at %.8x", msgContextP));

	newDestList = cniCreateMessage(&rc, msgContextP);
	if (rc != CCI_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to create new destination list "
								"(cniCreateMessage rc=%d)", rc));
		TRACE_RETURN(RSN_002, TT_FAILED_TO_CREATE_DESTINATION_LIST);
	}
	TRACE_MID(TRACE_INFO, ("New destination list created at %.8x", newDestList));

	/* Copy old destination list into new */
	rc = copyMessage(newDestList, destinationList);
	if (rc != TT_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to copy old destination list to new list"));
		TRACE_RETURN(RSN_003, TT_FAILED_TO_CREATE_DESTINATION_LIST);
	}

	/* address the Destination node, create it if it doesn't yet exist */
	destinationP = getMessageElement(newDestList,
		                             DESTINATION);
	if (!destinationP)
	{
		
		/* Get the root element of message */
		msgRootP = cniRootElement(&rc, newDestList);
		if (rc != CCI_SUCCESS)
		{
			TRACE_MID(TRACE_ERROR, ("cniRootElement for destination list FAILED"));
			TRACE_RETURN(RSN_004, TT_FAILED_TO_GET_MESSAGE_ROOT);
		}
		TRACE_MID(TRACE_INFO, ("Addressed destination list root element at %.8x", 
								msgRootP));

		destinationP = cniCreateElementAsFirstChild(&rc,
												   msgRootP);
		if (!destinationP)
		{
			TRACE_MID(TRACE_ERROR, ("Failed to create Destination child element "
									"(cniCreateElementAsFirstChild rc=%d)", rc));
			TRACE_RETURN(RSN_005, TT_FAILED_TO_CREATE_DESTINATION_LIST);
		}
		TRACE_MID(TRACE_INFO, ("Destination element created at %.8x", destinationP));

		cniSetElementName(&rc,
						  destinationP,
						  C.DESTINATION);
		if (rc != CCI_SUCCESS)
		{
			TRACE_MID(TRACE_ERROR, ("Failed to set Destination element name "
									"(cniSetElementName rc=%d)", rc));
			TRACE_RETURN(RSN_006, TT_FAILED_TO_CREATE_DESTINATION_LIST);
		}
		TRACE_MID(TRACE_INFO, ("Destination element name set to %S", L"Destination"));

		cniSetElementType(&rc,
						  destinationP,
						  CCI_ELEMENT_TYPE_NAME);
		if (rc != CCI_SUCCESS)
		{
			TRACE_MID(TRACE_ERROR, ("Failed to set Destination element type "
									"(cniSetElementType rc=%d)", rc));
			TRACE_RETURN(RSN_007, TT_FAILED_TO_CREATE_DESTINATION_LIST);
		}
		TRACE_MID(TRACE_INFO, ("Destination element type set"));
	}

	/* address the MQDestinationList node, create it if it doesn't yet exist */
	mqDestinationListP = getMessageElement(newDestList,
		                                   MQ_DESTINATION_LIST);
	if (!mqDestinationListP)
	{
		mqDestinationListP = cniCreateElementAsLastChild(&rc,
												         destinationP);
		if (rc != CCI_SUCCESS)
		{
			TRACE_MID(TRACE_ERROR, ("Failed to create MQDestinationList child element "
									"(cniCreateElementAsLastChild rc=%d)", rc));
			TRACE_RETURN(RSN_008, TT_FAILED_TO_CREATE_DESTINATION_LIST);
		}
		TRACE_MID(TRACE_INFO, ("MQDestinationList element created at %.8x", mqDestinationListP));

		cniSetElementName(&rc,
						  mqDestinationListP,
						  C.MQ_DESTINATION_LIST);
		if (rc != CCI_SUCCESS)
		{
			TRACE_MID(TRACE_ERROR, ("Failed to set MQDestinationList element name "
									"(cniSetElementName rc=%d)", rc));
			TRACE_RETURN(RSN_009, TT_FAILED_TO_CREATE_DESTINATION_LIST);
		}
		TRACE_MID(TRACE_INFO, ("MQDestinationList element name set to %S", C.MQ_DESTINATION_LIST));

		cniSetElementType(&rc,
						  mqDestinationListP,
						  CCI_ELEMENT_TYPE_NAME);
		if (rc != CCI_SUCCESS)
		{
			TRACE_MID(TRACE_ERROR, ("Failed to set MQDestinationList element type "
									"(cniSetElementType rc=%d)", rc));
			TRACE_RETURN(RSN_010, TT_FAILED_TO_CREATE_DESTINATION_LIST);
		}
		TRACE_MID(TRACE_INFO, ("MQDestinationList element type set"));
	}
    
	/* 
	 * create a new DestinationData element under the MQDestinationList node 
	 */
	destinationDataP = cniCreateElementAsLastChild(&rc,
								 			       mqDestinationListP);
	if (rc != CCI_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to create DestinationData child element "
								"(cniCreateElementAsLastChild rc=%d)", rc));
		TRACE_RETURN(RSN_011, TT_FAILED_TO_CREATE_DESTINATION_LIST);
	}
	TRACE_MID(TRACE_INFO, ("DestinationData element created at %.8x", destinationDataP));

	cniSetElementName(&rc,
					  destinationDataP,
					  C.DESTINATION_DATA);
	if (rc != CCI_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to set DestinationData element name "
								"(cniSetElementName rc=%d)", rc));
		TRACE_RETURN(RSN_012, TT_FAILED_TO_CREATE_DESTINATION_LIST);
	}
	TRACE_MID(TRACE_INFO, ("DestinationData element name set to %S", C.DESTINATION_DATA));

	cniSetElementType(&rc,
					  destinationDataP,
					  CCI_ELEMENT_TYPE_NAME);
	if (rc != CCI_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to set DestinationData element type "
								"(cniSetElementType rc=%d)", rc));
		TRACE_RETURN(RSN_014, TT_FAILED_TO_CREATE_DESTINATION_LIST);
	}
	TRACE_MID(TRACE_INFO, ("DestinationData element type set"));
	
	/*
	 * Add a queueManagerName value element to DestinationData
	 */
	qmnP = cniCreateElementAsLastChild(&rc,
					 			       destinationDataP);
	if (rc != CCI_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to create queueManagerName child element "
								"(cniCreateElementAsLastChild rc=%d)", rc));
		TRACE_RETURN(RSN_015, TT_FAILED_TO_CREATE_DESTINATION_LIST);
	}
	TRACE_MID(TRACE_INFO, ("queueManagerName element created at %.8x", qmnP));

	cniSetElementName(&rc,
					  qmnP,
					  C.QUEUE_MANAGER_NAME);
	if (rc != CCI_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to set queueManagerName element name "
								"(cniSetElementName rc=%d)", rc));
		TRACE_RETURN(RSN_016, TT_FAILED_TO_CREATE_DESTINATION_LIST);
	}
	TRACE_MID(TRACE_INFO, ("queueManagerName element name set to %S", L"queueManagerName"));

	cniSetElementType(&rc,
					  qmnP,
					  CCI_ELEMENT_TYPE_NAME_VALUE);
	if (rc != CCI_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to set queueManagerName element type "
								"(cniSetElementType rc=%d)", rc));
		TRACE_RETURN(RSN_017, TT_FAILED_TO_CREATE_DESTINATION_LIST);
	}
	TRACE_MID(TRACE_INFO, ("queueManagerName element type set"));
	
	cniSetElementCharacterValue(&rc,
		                        qmnP,
								queueManagerName,
								MQ_Q_MGR_NAME_LENGTH);
	if (rc != CCI_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to set queueManagerName element value "
			                    "(cniSetElementType rc=%d)", rc));
		TRACE_RETURN(RSN_018, TT_FAILED_TO_CREATE_DESTINATION_LIST);
	}
	TRACE_MID(TRACE_INFO, ("queueManagerName element value set to '%S'",
		                    queueManagerName));

	/*
	 * Add a queueName value element to DestinationData
	 */
	qnP = cniCreateElementAsLastChild(&rc,
				 			          destinationDataP);
	if (rc != CCI_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to create queueName child element "
								"(cniCreateElementAsLastChild rc=%d)", rc));
		TRACE_RETURN(RSN_019, TT_FAILED_TO_CREATE_DESTINATION_LIST);
	}
	TRACE_MID(TRACE_INFO, ("queueName element created at %.8x", qnP));

	cniSetElementName(&rc,
					  qnP,
					  C.QUEUE_NAME);
	if (rc != CCI_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to set queueName element name "
								"(cniSetElementName rc=%d)", rc));
		TRACE_RETURN(RSN_020, TT_FAILED_TO_CREATE_DESTINATION_LIST);
	}
	TRACE_MID(TRACE_INFO, ("queueName element name set to %S", C.QUEUE_NAME));

	cniSetElementType(&rc,
					  qnP,
					  CCI_ELEMENT_TYPE_NAME_VALUE);
	if (rc != CCI_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to set queueName element type "
								"(cniSetElementType rc=%d)", rc));
		TRACE_RETURN(RSN_021, TT_FAILED_TO_CREATE_DESTINATION_LIST);
	}
	TRACE_MID(TRACE_INFO, ("queueName element type set"));
	
	cniSetElementCharacterValue(&rc,
		                        qnP,
								queueName,
								MQ_Q_NAME_LENGTH);
	if (rc != CCI_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to set queueName element value "
			                    "(cniSetElementType rc=%d)", rc));
		TRACE_RETURN(RSN_022, TT_FAILED_TO_CREATE_DESTINATION_LIST);
	}
	TRACE_MID(TRACE_INFO, ("queueName element value set to '%S'",
		                    queueName));

    *newDestinationListP = newDestList;

	TRACE_RETURN(RSN_023, TT_SUCCESS);
}


