/******************************************************************************/
/*                                                                            */
/* Licensed Materials - Property of IBM                                       */
/* (C) Copyright IBM Corp. 2000                                               */
/*                                                                            */
/* Program name    : log.c                                                    */
/* Output name     : none                                                     */
/*                                                                            */
/* Compile options : /VERBOSE /LD /MT                                         */
/*                   INCLUDE must contain <install dir>\inlude\plugin         */
/* Link options    : /DLL                                                     */
/*                   LIB must contain <install dir>\bin\imbdfplg.lib          */
/*                                                                            */
/* Description     : Contains various logging functions for persisting the    */
/*                   state of the aggregator data.                            */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/* Version 1.0                                                                */
/* Author: Mike Brady                    Date: May 2001                       */
/*                                                                            */
/******************************************************************************/

/* Include Headers List */
#include "aggregator.h"      /* Specific header for this dll                       */
#include <malloc.h>
#include <memory.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "log.h"

#include <BipCni.h>		/* Function prototypes for Plug-in Node functions     */

int getLogQueueDetails(IN queueDetails_t **queueDetailsPP,
					   OUT char **errorMessageP);
int writeLog(IN char *logMessage,
			 IN int messageLength,
			 IN char *aggregateId,
	 		 OUT char **errorStringP);

/******************************************************************************/
/* logNewInstance                                                             */
/******************************************************************************/
int logNewInstance(IN aggregateInstance_t *aggregateInstanceP,
	 		       OUT char **errorStringP)
{
	logRecordHeader_t recordHeader;
	int msgFlowNameLength,
		messageLength,
		aggNameLength;
	char logMessage[sizeof(logRecordHeader_t) + 
		            sizeof(int) +
		            sizeof(CciChar) * (MAX_STRING_SIZE + 1)],
		 *messageP;

	int retCode;

    TRACE_ENTRY("logNewInstance", ("aggregateInstanceP=%.8x", aggregateInstanceP));

	/* build a new logRecord header */
	memset(&recordHeader, 0, sizeof(recordHeader));
	recordHeader.recordType = NEW_INSTANCE;
	memcpy(recordHeader.aggregateId, aggregateInstanceP->aggregateId,
		   MQ_CORREL_ID_LENGTH);

	/* build a new instance log record */

/*
 * | logRecordHeader_t | msgFlowNameLength | msgFlowName | aggNameLength | aggName |
 */

	/* finally, build the log message */
    messageP = logMessage;

	/* logRecordHeader_t */
	memcpy(messageP, &recordHeader, sizeof(logRecordHeader_t));
	messageP += sizeof(logRecordHeader_t);

	/* msgFlowNameLength + msgFlowName */
	msgFlowNameLength = (UClen(aggregateInstanceP->deployedAggregateP->contextP->messageFlowName) + 1) * 
		          sizeof(CciChar);
	memcpy(messageP, &msgFlowNameLength, sizeof(msgFlowNameLength));
	messageP += sizeof(msgFlowNameLength);

	UCcpy((CciChar *)messageP, aggregateInstanceP->deployedAggregateP->contextP->messageFlowName);
	messageP += msgFlowNameLength;

	/* aggNameLength + aggName */
	aggNameLength = (UClen(aggregateInstanceP->deployedAggregateP->name) + 1) * 
		          sizeof(CciChar);
	memcpy(messageP, &aggNameLength, sizeof(aggNameLength));
	messageP += sizeof(aggNameLength);

	UCcpy((CciChar *)messageP, aggregateInstanceP->deployedAggregateP->name);
	messageP += aggNameLength;


	/* calculate the length of the message */
	messageLength = (int)messageP - (int)logMessage;

	retCode = TT_SUCCESS;
    retCode = writeLog(logMessage,
		               messageLength,
		               aggregateInstanceP->aggregateId,
					   errorStringP);

    TRACE_RETURN(RSN_001, retCode);
}

/******************************************************************************/
/* logMember                                                                  */
/******************************************************************************/
int logMember(IN aggregateMember_t *memberP,
	 		  OUT char **errorStringP)
{
    aggregateMemberLoggedData_t *memberLoggedDataP;
	logRecordHeader_t recordHeader;
	logMember_t logRecord;
	int nameLength,
		messageLength;
	char logMessage[sizeof(logRecordHeader_t) + 
		            sizeof(int) +
		            sizeof(CciChar) * (MAX_STRING_SIZE + 1) +
					sizeof(logMember_t)],
		 *messageP;

	int retCode;

    TRACE_ENTRY("logMember", ("memberP=%.8x", memberP));

	/* build a new logRecord header */
	memset(&recordHeader, 0, sizeof(recordHeader));
	recordHeader.recordType = MEMBER;
	memcpy(recordHeader.aggregateId, 
		   memberP->aggregateInstanceP->aggregateId, MQ_CORREL_ID_LENGTH);

	/* build a new log member record */
	memberLoggedDataP = (aggregateMemberLoggedData_t *)memberP;
	memset(&logRecord, 0, sizeof(logRecord));
    memcpy(&logRecord.instanceTimeoutTime, 
		   &memberP->aggregateInstanceP->timeoutTime, sizeof(time_t));
	memcpy(&logRecord.loggedData, memberLoggedDataP, sizeof(aggregateMemberLoggedData_t));

/*
 * | logRecordHeader_t | aggNameLength | aggName | logMember_t |
 */

	/* finally, build the log message */
    messageP = logMessage;

	/* logRecordHeader_t */
	memcpy(messageP, &recordHeader, sizeof(logRecordHeader_t));
	messageP += sizeof(logRecordHeader_t);

	/* aggNameLength + aggName */
	nameLength = (UClen(memberP->aggregateInstanceP->deployedAggregateP->name) + 1) * 
		          sizeof(CciChar);
	memcpy(messageP, &nameLength, sizeof(nameLength));
	messageP += sizeof(nameLength);

	UCcpy((CciChar *)messageP, memberP->aggregateInstanceP->deployedAggregateP->name);
	messageP += nameLength;

	/* logMember_t */
	memcpy(messageP, &logRecord, sizeof(logMember_t));
	messageP += sizeof(logMember_t);

	/* calculate the length of the message */
	messageLength = (int)messageP - (int)logMessage;

	retCode = TT_SUCCESS;
    retCode = writeLog(logMessage,
		               messageLength,
		               memberP->aggregateInstanceP->aggregateId,
					   errorStringP);

    TRACE_RETURN(RSN_001, retCode);
}

/******************************************************************************/
/* logReply                                                                   */
/******************************************************************************/
int logReply(IN aggregateMember_t *memberP,
	 		 OUT char **errorStringP)
{
    aggregateReplyLoggedData_t *replyLoggedDataP;
	logRecordHeader_t recordHeader;
	logReply_t logRecord;
	char *logMessage,
		 *messageP;
	int retCode,
		msgLength,
		nameLength;

    TRACE_ENTRY("logReply", ("memberP=%.8x", memberP));

	/* build a new logRecord header */
	memset(&recordHeader, 0, sizeof(recordHeader));
	recordHeader.recordType = REPLY;
	memcpy(recordHeader.aggregateId, 
		   memberP->aggregateInstanceP->aggregateId, MQ_CORREL_ID_LENGTH);

	/* build a new log reply record */
	replyLoggedDataP = (aggregateReplyLoggedData_t *)memberP;

	memset(&logRecord, 0, sizeof(logRecord));
    memcpy(&logRecord.md, memberP->mqmdP, sizeof(MQMD));
	memcpy(&logRecord.loggedData, replyLoggedDataP, sizeof(aggregateReplyLoggedData_t));

/*
 * | logRecordHeader_t | aggNameLength | aggName | logReply_t | messageBody |
 */

	/* finally, build the log message */
	/* allocate a log record large enough to contain the reply message body */
	logMessage = (char *)malloc(sizeof(logRecordHeader_t) +
		                        sizeof(int) +
		                        sizeof(CciChar) * (MAX_STRING_SIZE + 1) +
		                        sizeof(logReply_t) +
		                        memberP->messageLength);
	TRACE_ALLOC("log.c-1", logMessage);

    messageP = logMessage;

	/* logRecordHeader_t */
	memcpy(messageP, &recordHeader, sizeof(logRecordHeader_t));
	messageP += sizeof(logRecordHeader_t);

	/* aggNameLength + aggName */
	nameLength = (UClen(memberP->aggregateInstanceP->deployedAggregateP->name) + 1) * 
		          sizeof(CciChar);
	memcpy(messageP, &nameLength, sizeof(nameLength));
	messageP += sizeof(nameLength);

	UCcpy((CciChar *)messageP, memberP->aggregateInstanceP->deployedAggregateP->name);
	messageP += nameLength;

	/* logReply_t */
	memcpy(messageP, &logRecord, sizeof(logReply_t));
	messageP += sizeof(logReply_t);

	/* messageBody */
	memcpy(messageP, 
		   memberP->messageP, memberP->messageLength);
	messageP += memberP->messageLength;

	/* calculate the length of the message */
	msgLength = (int)messageP - (int)logMessage;

	retCode = TT_SUCCESS;
    retCode = writeLog(logMessage,
		               msgLength,
		               memberP->aggregateInstanceP->aggregateId,
					   errorStringP);

	/* free up the log message */
	free(logMessage);
	TRACE_FREE("log.c-14", logMessage);

    TRACE_RETURN(RSN_001, retCode);
}

/******************************************************************************/
/* logFinish                                                                  */
/******************************************************************************/
int logFinish(IN char *aggregateId,
			  IN CciChar *aggregateName,
	  		  OUT char **errorStringP)
{
	logRecordHeader_t recordHeader;
	int messageLength;
	char logMessage[sizeof(logRecordHeader_t) + 
		            sizeof(int) +
		            sizeof(CciChar) * (MAX_STRING_SIZE + 1)],
		 *messageP;

	int retCode,
		nameLength;

    TRACE_ENTRY("logFinish", ("aggregateId=%.8x, aggregateName=%S", 
		                      aggregateId, aggregateName));

/*
 * | logRecordHeader_t | aggNameLength | aggName |
 */

	/* build a new logRecord header */
	memset(&recordHeader, 0, sizeof(recordHeader));
	recordHeader.recordType = FINISH;
	memcpy(recordHeader.aggregateId, aggregateId,MQ_CORREL_ID_LENGTH);

	/* finally, build the log message */
    messageP = logMessage;

	/* logRecordHeader_t */
	memcpy(messageP, &recordHeader, sizeof(logRecordHeader_t));
	messageP += sizeof(logRecordHeader_t);

	/* aggNameLength + aggName */
	nameLength = (UClen(aggregateName) + 1) * sizeof(CciChar);
	memcpy(messageP, &nameLength, sizeof(nameLength));
	messageP += sizeof(nameLength);

	UCcpy((CciChar *)messageP, aggregateName);
	messageP += nameLength;

	/* calculate the length of the message */
	messageLength = (int)messageP - (int)logMessage;

	retCode = TT_SUCCESS;
    retCode = writeLog(logMessage,
		               messageLength,
		               aggregateId,
					   errorStringP);

    TRACE_RETURN(RSN_001, retCode);
}


typedef struct msgFlowAggregate
{
    char aggregateId[MQ_CORREL_ID_LENGTH];
	bool_t membersExist;
} msgFlowAggregate_t;

/******************************************************************************/
/* logRecoverInstances                                                        */
/******************************************************************************/
int logRecoverInstances(IN context_t *contextP,
                        IN deployedAggregate_t *deployedAggregateP,
						OUT bool_t *instancesRecoveredP,
						OUT char **errorStringP)
{

    MQLONG   compCode,                  /* completion code               */
             reason;                    /* reason code                   */
    MQOD od = {MQOD_DEFAULT};           /* Object Descriptor */
    MQMD md,                            /* Message Descriptor */
         mdDefault = {MQMD_DEFAULT};    /* Message Descriptor default values */
    MQGMO gmo,                          /* get message options */
          gmoDefault = {MQGMO_DEFAULT}; /* get message options default values */
	char correlId[MQ_CORREL_ID_LENGTH],
		 *logRecord,
		 *msgBody,
		 errorBuffer[1000],
		 buffer[100];
	CciChar *aggregateName;
	logRecordHeader_t *logRecordHeaderP;
    logMember_t *logMemberP;
	logReply_t *logReplyP;
	int retCode,
		nameLength,
		numInstancesRecovered=0;
	long messageLength,
		 bufferSize;
	queueDetails_t *queueDetailsP;
	bool_t moreLogMessages,
		   found;
    aggregateInstance_t *currentAggregateP,
	                    *aggregateAnchorP;
	aggregateMember_t *memberP,
		              *nextMemberP;
    aggregateMemberLoggedData_t *memberLoggedDataP;
    newInstanceRecord_t *newInstanceRecordP;
	msgIdIndexElement_t *msgIndexP;
	msgFlowAggregate_t *validAggregateIdP = NULL,
		               *currentMsgFlowAggregateP;
	int numValidAggregateIds = 0,
	    numSpareValidAggregateIds = 0,
		i;

    TRACE_ENTRY("logRecoverInstances", ("contextP=%.8x, deployedAggregateP=%.8x", 
		                                 contextP, deployedAggregateP));

	/*
	 * Assumption:
	 *
	 * On entry to this function, the caller has got the aggregate chain locked
	 */
	/* 
	 * This function is called during aggregator initialisation.  It is called once per
	 * deployed aggregate.  Log records that are relevant to the deployed aggregate must
	 * have the following characteristics.
	 *
	 *  - the correlation id value must match the current Execution Group.  This is to 
	 *    ensure that we only recover those resources which belong to the current Execution
	 *    Group.  After all, it would be possible to have the same deployed aggregate exist
	 *    in more than one Execution Group.
	 *
	 *  - the aggregate name from the log record must match the specified deployed 
	 *    aggregate name.
	 *
	 *  - the message flow name from the log record must match the current message flow.
	 */

	if (errorStringP)
	{
		*errorStringP = "";
	}

	*instancesRecoveredP = FALSE;
	retCode = getLogQueueDetails(&queueDetailsP, errorStringP);

	if (retCode != TT_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to get recovery queue details"));
		TRACE_RETURN(RSN_005, TT_FAILED_TO_ACCESS_RECOVERY_MEDIA);
	}

	/* We're only interested in records belonging to the current Execution Group */
	memset(correlId, '\0', MQ_CORREL_ID_LENGTH);
	if (strlen(executionGroupName) > MQ_CORREL_ID_LENGTH)
	{
		/* take the first and last 12 chars of execution group name */
		memcpy(correlId, executionGroupName, 
			   MQ_CORREL_ID_LENGTH/2);
		memcpy(correlId+MQ_CORREL_ID_LENGTH/2, 
			   executionGroupName+strlen(executionGroupName)-MQ_CORREL_ID_LENGTH/2, 
			   MQ_CORREL_ID_LENGTH/2);
	}
	else
	{
		memcpy(correlId, executionGroupName, strlen(executionGroupName));
	}

	/* don't wait for messages */
	gmoDefault.Options += MQGMO_NO_WAIT;

	memcpy(&gmo, &gmoDefault, sizeof(gmo));

	/* first time through, start at first message in the recovery queue */
	gmo.Options += MQGMO_BROWSE_FIRST;

	/* make sure we move on to next message on subsequent queue reads */
	gmoDefault.Options += MQGMO_BROWSE_NEXT;

	/* set the correlation id to the execution group name */
	memcpy(mdDefault.CorrelId, correlId, MQ_CORREL_ID_LENGTH);
	memcpy(&md, &mdDefault, sizeof(md));
	TRACE_MID(TRACE_INFO, ("Retrieving messages with CorrelId value..."));
    TRACE_MIDHEX(TRACE_INFO, mdDefault.CorrelId, MQ_CORREL_ID_LENGTH);

	/* allocate the initial buffer size */
	logRecord = (char *)malloc(DEFAULT_LOG_RECORD_SIZE);
	TRACE_ALLOC("log.c-2", logRecord);
	bufferSize = DEFAULT_LOG_RECORD_SIZE;

	/* read all messages from the queue which belong to this aggregate */
	moreLogMessages = TRUE;
	while(moreLogMessages)
	{
		MQGET(queueDetailsP->qmConn,        /* connection handle */
			  queueDetailsP->queueHandle,   /* object handle */
			  &md,                          /* message descriptor */
			  &gmo,                         /* get message options */
			  bufferSize,                   /* size of receive buffer */
	          logRecord,                    /* receive buffer */
	  (long *)&messageLength,               /* returned message length */
			  &compCode,                    /* completion code */
			  &reason);                     /* reason code */

		if (reason == MQRC_NO_MSG_AVAILABLE)
		{
			TRACE_MID(TRACE_INFO, ("No more messages available"));
			moreLogMessages = FALSE;
			continue;
		}
 
		/* increase buffer if too small */
		if (reason == MQRC_TRUNCATED_MSG_FAILED)
		{
			TRACE_MID(TRACE_INFO, ("Message truncated, increasing buffer from %d to %d "
				                   "bytes", bufferSize, messageLength));
			logRecord = (char *)realloc(logRecord, messageLength);
			TRACE_ALLOC("log.c-3", logRecord);
			continue;
		}
 
		if (reason != MQRC_NONE)
		{
			sprintf(errorBuffer, "MQGET for '%s' failed with "
								  "reason code %ld",
								  AGGREGATOR_RECOVERY_QUEUE, reason);
			TRACE_MID(TRACE_ERROR, ("%s", errorBuffer));
			free(logRecord);
			TRACE_FREE("log.c-15", logRecord);

			if (validAggregateIdP)
			{
				free(validAggregateIdP);
				TRACE_FREE("log.c-17", validAggregateIdP);
			}
			if (errorStringP)
			{
				*errorStringP = (char *)malloc(strlen(errorBuffer) + 1);
				TRACE_ALLOC("log.c-4", errorStringP);
				strcpy(*errorStringP, errorBuffer);
			}
			TRACE_RETURN(RSN_005, TT_FAILED_TO_READ_LOG_RECORD);
		}


		/* make sure we get the next record on the next queue read */
		memcpy(&md, &mdDefault, sizeof(md));
		memcpy(&gmo, &gmoDefault, sizeof(gmo));

		/* check the record type */
		logRecordHeaderP = (logRecordHeader_t *)logRecord;
	
		/* process the log record just read */
		TRACE_MID(TRACE_INFO, ("Log message read from queue, type is %s",
				               logRecordHeaderP->recordType == NEW_INSTANCE ? "NEW_INSTANCE" :
				               logRecordHeaderP->recordType == MEMBER ? "MEMBER" :
			                   logRecordHeaderP->recordType == REPLY ? "REPLY" :
			                   logRecordHeaderP->recordType == FINISH ? "FINISH" :
                               "'unknown'"));

		/* make a note of the aggregate id if this is a new instance record */
		if (logRecordHeaderP->recordType == NEW_INSTANCE)
		{
            /*
			 * | logRecordHeader_t | msgFlowNameLength | msgFlowName | aggNameLength | aggName |
			 */
			newInstanceRecordP = (newInstanceRecord_t *)
				                  (logRecord + sizeof(logRecordHeader_t));
			
			/* ignore records which do not belong to the current Message Flow */
		    if (UCcmp(contextP->messageFlowName, newInstanceRecordP->messageFlowName))
			{
				/* message flow names don't match, ignore record */
				TRACE_MID(TRACE_INFO, ("Message flow name %S does not match current "
					                   "message flow %S, ignoring log record", 
									   newInstanceRecordP->messageFlowName,
									   contextP->messageFlowName));
				continue;
			}
			
			/* make sure there's a spare slot in the list of valid ids */
			if (!numSpareValidAggregateIds)
			{
				if (validAggregateIdP)
				{
					validAggregateIdP = (msgFlowAggregate_t *)realloc(validAggregateIdP, 
						           sizeof(msgFlowAggregate_t) * 
								        (numSpareValidAggregateIds + 10));
					TRACE_ALLOC("log.c-5", validAggregateIdP);
					TRACE_MID(TRACE_INFO, ("Increasing slots in aggregate id table to %d",
						                   numSpareValidAggregateIds+10));
				}
				else
				{
					validAggregateIdP = (msgFlowAggregate_t *)malloc(sizeof(msgFlowAggregate_t) * 10);
					TRACE_ALLOC("log.c-6", validAggregateIdP);
					TRACE_MID(TRACE_INFO, ("Allocating %d slots in aggregate id table",
						                   numSpareValidAggregateIds+10));
				}
				numSpareValidAggregateIds += 10;
			}

			/* add aggregate id to list of valid ids */
			memcpy((validAggregateIdP+
				            (numValidAggregateIds*sizeof(msgFlowAggregate_t)))->aggregateId, 
				   logRecordHeaderP->aggregateId, MQ_CORREL_ID_LENGTH);
			(validAggregateIdP+
				  (numValidAggregateIds*sizeof(msgFlowAggregate_t)))->membersExist = FALSE; 
			numValidAggregateIds++;
			numSpareValidAggregateIds--;
			continue;
		}

		/* 
		 * This is a MEMBER, REPLY or FINISH record.  Check the aggregate id of this 
		 * record against our list of valid ids.  If it's not listed, then ignore it.
		 */
		found = FALSE;
		for (i=0; i<numValidAggregateIds; i++)
		{
			currentMsgFlowAggregateP = validAggregateIdP + 
				                      (i * sizeof(msgFlowAggregate_t));
			if (!memcmp(logRecordHeaderP->aggregateId, 
				        currentMsgFlowAggregateP->aggregateId, MQ_CORREL_ID_LENGTH))
			{
				found = TRUE;
				TRACE_MID(TRACE_INFO, ("Matching aggregate id found in "
					                   "valid aggregate id table"));
				break;
			}
		}

        if (!found)
		{
			/* ignore log records which do not belong to a valid aggregate */
			TRACE_MID(TRACE_INFO, ("Matching aggregate id not found - ignoring record"));
			continue;
		}

		/* process matching log record */
		TRACE_MID(TRACE_INFO, ("Matching aggregate log record found"));

		/*
		 * If this is a MEMBER record, then this may be the first member of an
		 * aggregate instance.  In this case we must create the aggregate, otherwise
		 * we just add the member to the existing aggregate.
		 *
		 * If this is a REPLY record, then we expect to find an existing aggregate
		 * to add the reply details to, otherwise we assume the recovery log is
		 * corrupted and raise an exception.
		 *
		 * If this is a FINISH record, then we remove all control blocks related to the
		 * aggregate id.
         */

		/* if aggregate instance doesn't exist, create it */

		/* address the aggregate instance chain */
		currentAggregateP = findAggregateInstanceFromContext(contextP,
 			                               logRecordHeaderP->aggregateId);
		TRACE_MID(TRACE_INFO, ("currentAggregatP value is %.8x", currentAggregateP));

		/* locate the aggregate name portion of the log record */
		nameLength = *(int *)(logRecord + sizeof(logRecordHeader_t));
		TRACE_MID(TRACE_INFO, ("aggregateNameLength is %d", nameLength));
		aggregateName = (CciChar *)(logRecord + sizeof(logRecordHeader_t) + sizeof(int));

		if (logRecordHeaderP->recordType == MEMBER)
		{
			/*
			 * | logRecordHeader_t | aggNameLength | aggName | logMember_t |
			 */

			/* locate the message record portion of the log record */
			logMemberP = (logMember_t *)((char *)aggregateName + nameLength);
			memberLoggedDataP = &logMemberP->loggedData;

			if (!currentAggregateP)
			{

				/* this must be the first MEMBER record for an aggregate instance */

			    if (currentMsgFlowAggregateP->membersExist)
				{
					/* error... */
				}

				/* aggregate instance doesn't exist yet, create it */
				currentAggregateP = (aggregateInstance_t *)malloc(sizeof(aggregateInstance_t));
				memset(currentAggregateP, 0, sizeof(aggregateInstance_t));

				TRACE_MID(TRACE_INFO, ("Created new aggregate instance at %.8x",
										currentAggregateP));
				TRACE_ALLOC("log.c-7", currentAggregateP);

				/* initialise the aggregate id */
				memcpy(currentAggregateP->aggregateId, logRecordHeaderP->aggregateId, 
					   MQ_CORREL_ID_LENGTH);

				/* initialise pointer back to deployed aggregate */
				currentAggregateP->deployedAggregateP = 
										   (deployedAggregate_t *)(contextP->nodeDataP);

				/* initialise the common message id */
				memcpy(currentAggregateP->initialMsgId, 
					   memberLoggedDataP->originalMessageId, 
					   MQ_MSG_ID_LENGTH);
				currentAggregateP->commonMsgId = TRUE;

				/* add to the aggregate instance chain */
				aggregateAnchorP = &(currentAggregateP->deployedAggregateP->aggregateAnchor);

				INSERT_AFTER_IN_LINKED_LIST(aggregateAnchorP, currentAggregateP);
	            
				/* let caller know that at least one instance has been recovered */
				numInstancesRecovered++;
			}

			/* create a new aggregate member */
			memberP = (aggregateMember_t *)malloc(sizeof(aggregateMember_t));
			memset(memberP, 0, sizeof(aggregateMember_t));
			TRACE_MID(TRACE_INFO, ("New aggregate member created at %.8x",
								   memberP));
			TRACE_ALLOC("log.c-8", memberP);

			memcpy(memberP, memberLoggedDataP, sizeof(aggregateMemberLoggedData_t));
			memberP->mqmdP = NULL;
			memberP->messageP = NULL;
			memberP->messageLength = 0;

#if 0
if (!addrListP)
{
	addrListP = (addressList_t *)malloc(sizeof(addressList_t) + (100-1)*sizeof(void *));
	TRACE_ALLOC("log.c-9", addrListP);
	addrListP->nextAddress = 0;
	addrListP->numAddresses = 100;
}

addrListP->address[addrListP->nextAddress] = (void *)(memberP);
addrListP->nextAddress++;

testAddress(1);
#endif
			currentAggregateP->memberCount++;
			currentAggregateP->outstandingReplies++;
            currentMsgFlowAggregateP->membersExist = TRUE;

			TRACE_MID(TRACE_INFO, ("Aggregate descriptor member count incremented to %d",
								   currentAggregateP->memberCount));
			TRACE_MID(TRACE_INFO, ("Aggregate descriptor outstanding replies incremented to %d",
								   currentAggregateP->outstandingReplies));

			memberP->aggregateInstanceP = currentAggregateP;
			TRACE_MID(TRACE_INFO, ("Parent aggregate descriptor set to %.8x",
								   memberP->aggregateInstanceP));

			/* update the aggregate timeout value */
			memcpy(&currentAggregateP->timeoutTime, 
				   &logMemberP->instanceTimeoutTime,
				   sizeof(currentAggregateP->timeoutTime));

			/* add the new member to the aggregate instance, in sequence number order */
			addMemberToAggregateInstance(memberP,
										 currentAggregateP);

			/* add message to message id index */
		    dispHex((char *)(memberP->newMessageId), MQ_MSG_ID_LENGTH, buffer);
			TRACE_MID(TRACE_INFO, ("Updating msgId value '%s'", buffer));
			msgIndexP = updateMsgIdIndex(memberP->newMessageId, &msgIdIndexAnchor);
			msgIndexP->memberP = memberP;
			memberP->indexP = msgIndexP;

			/* reset the aggregate reply timeout */
			setAggregateReplyAlarm(currentAggregateP,
								   memberP,
								   TRUE);  /* recovery processing in effect */
		}
		else
		if (logRecordHeaderP->recordType == REPLY)
		{
			/* processing a REPLY type log message */

			/*
			 * | logRecordHeader_t | aggNameLength | aggName | logReply_t | messageBody |
			 */

			if (!currentAggregateP)
			{
				UCsprintf(errorBuffer, "Unexpected REPLY record encountered in log for "
									 "aggregate '%S'.  Log is assumed to be corrupt.",
								 	 aggregateName);
				/* gen error message */
				TRACE_MID(TRACE_ERROR, ("%s", errorBuffer));

				if (validAggregateIdP)
				{
					free(validAggregateIdP);
					TRACE_FREE("log.c-17", validAggregateIdP);
				}
				if (errorStringP)
				{
					*errorStringP = (char *)malloc(strlen(errorBuffer) + 1);
					TRACE_ALLOC("log.c-10", errorStringP);
					strcpy(*errorStringP, errorBuffer);
				}
				TRACE_RETURN(RSN_005, TT_UNEXPECTED_REPLY_MESSAGE_FOUND_IN_LOG);
			}

			/* address the logged reply data */
		    logReplyP = (logReply_t *)((char *)aggregateName + nameLength);

		    dispHex((char *)(logReplyP->md.CorrelId), MQ_MSG_ID_LENGTH, buffer);
			TRACE_MID(TRACE_INFO, ("Looking for correlId value '%s'", buffer));
			/* find the request message corresponding to this reply */
			msgIndexP = findMsgIndexElement((char *)logReplyP->md.CorrelId, 
				                             &msgIdIndexAnchor);

			if (!msgIndexP)
			{
				/* we always expect to find an aggregate instance, else error */
				UCsprintf(errorBuffer, "No matching request message found for recovered "
					                    "log reply message for aggregate '%S'.  "
										"Log is assumed to be corrupt.",
										aggregateName);
				TRACE_MID(TRACE_ERROR, ("%s", errorBuffer));
				if (validAggregateIdP)
				{
					free(validAggregateIdP);
					TRACE_FREE("log.c-17", validAggregateIdP);
				}
				if (errorStringP)
				{
					*errorStringP = (char *)malloc(strlen(errorBuffer) + 1);
					TRACE_ALLOC("log.c-11", errorStringP);
					strcpy(*errorStringP, errorBuffer);
				}
				TRACE_RETURN(RSN_005, TT_RECOVERED_REPLY_MESSAGE_MISSING_REQUEST);
			}

			/* message reply belongs to an active aggregate */
			memberP = msgIndexP->memberP;
			TRACE_MID(TRACE_INFO, ("Processing aggregate member at %.8x", memberP));

			/* add reply to aggregate */
			currentAggregateP->outstandingReplies--;
			TRACE_MID(TRACE_INFO, ("Outstanding reply count decremented to %d",
			               currentAggregateP->outstandingReplies));
            memberP->mqmdP = (char *)malloc(sizeof(MQMD));
			TRACE_ALLOC("log.c-12", memberP->mqmdP);
			memcpy(memberP->mqmdP, &logReplyP->md, sizeof(MQMD));

			/* copy logged details into aggregate member structure */
			memcpy(memberP, &logReplyP->loggedData, 
				   sizeof(aggregateReplyLoggedData_t));

			if (memberP->messageLength)
			{
				msgBody = (char *)logReplyP + sizeof(logReply_t);
				memberP->messageP = (char *)malloc(memberP->messageLength);
				memcpy(memberP->messageP, msgBody, memberP->messageLength);
				TRACE_MID(TRACE_INFO, ("Message payload stored at %.8x, length %d",
									   memberP->messageP, memberP->messageLength));
				TRACE_ALLOC("log.c-13", memberP->messageP);
			}
			else
			{
				TRACE_MID(TRACE_INFO, ("Empty message body"));
			}
		}
		else
		if (logRecordHeaderP->recordType == FINISH)
		{
			/* processing a FINISH type log message */

			/*
			 * | logRecordHeader_t | aggNameLength | aggName |
			 */

			if (!currentAggregateP)
			{
				UCsprintf(errorBuffer, "Unexpected FINISH record encountered in log for "
									 "aggregate '%S'.  Log is assumed to be corrupt.",
								 	 aggregateName);
				/* gen error message */
				TRACE_MID(TRACE_ERROR, ("%s", errorBuffer));

				if (validAggregateIdP)
				{
					free(validAggregateIdP);
					TRACE_FREE("log.c-17", validAggregateIdP);
				}
				if (errorStringP)
				{
					*errorStringP = (char *)malloc(strlen(errorBuffer) + 1);
					TRACE_ALLOC("log.c-14", errorStringP);
					strcpy(*errorStringP, errorBuffer);
				}
				TRACE_RETURN(RSN_005, TT_UNEXPECTED_FINISH_MESSAGE_FOUND_IN_LOG);
			}

			/* remove all control blocks for this aggregate instance */
			REMOVE_FROM_LINKED_LIST(currentAggregateP);
			REMOVE_FROM_ALARM_CHAIN(currentAggregateP);

			/* free up all aggregate members */
			memberP = currentAggregateP->memberAnchor.nextP;

			while(memberP)
			{
				nextMemberP = memberP->nextP;

				TRACE_MID(TRACE_INFO, ("processing aggregate member at %.8x",
									   memberP));

				if (memberP->messageP)
				{
					TRACE_MID(TRACE_INFO, ("Releasing message storage at %.8x",
											memberP->messageP));
					free(memberP->messageP);
					TRACE_FREE("log.c-2", memberP->messageP);
				}

				if (memberP->mqmdP)
				{
					TRACE_MID(TRACE_INFO, ("Releasing MQMD storage at %.8x",
											memberP->mqmdP));
					free(memberP->mqmdP);
					TRACE_FREE("log.c-23", memberP->mqmdP);
				}

				TRACE_MID(TRACE_INFO, ("Releasing member storage at %.8x",
									   memberP));
				free(memberP);
				TRACE_FREE("log.c-24", memberP);

				memberP = nextMemberP;
			}

			TRACE_MID(TRACE_INFO, ("Releasing aggregate descriptor storage at %.8x",
								   currentAggregateP));
			free(currentAggregateP);
			TRACE_FREE("log.c-25", currentAggregateP);
			numInstancesRecovered--;
		}
		else
		{
			/* unknown log record type */
		}
	}

	free(logRecord);
	TRACE_FREE("log.c-15", logRecord);

	if (validAggregateIdP)
	{
		free(validAggregateIdP);
		TRACE_FREE("log.c-17", validAggregateIdP);
	}
	if (numInstancesRecovered)
	{
		TRACE_MID(TRACE_INFO, ("%d aggregate instances recovered, setting "
			                   "instancesRecovered to TRUE", numInstancesRecovered));
		*instancesRecoveredP = TRUE;
	}
	else
	{
		TRACE_MID(TRACE_INFO, ("No aggregate instances recovered, setting "
			                   "instancesRecovered to FALSE"));
		*instancesRecoveredP = FALSE;
	}
    TRACE_RETURN(RSN_001, TT_SUCCESS);
}

/******************************************************************************/
/* logRemoveEntries                                                           */
/******************************************************************************/
int logRemoveEntries(IN char *aggregateId,
					 IN bool_t commit,
			 		 OUT char **errorStringP)
{

    MQLONG   compCode,                  /* completion code               */
             reason;                    /* reason code                   */
    MQOD od = {MQOD_DEFAULT};           /* Object Descriptor */
    MQMD md = {MQMD_DEFAULT};           /* Message Descriptor */
    MQGMO gmo = {MQGMO_DEFAULT};        /* get message options */
	char *queueNameP = AGGREGATOR_RECOVERY_QUEUE,
		 correlId[MQ_CORREL_ID_LENGTH];
	logRecordHeader_t logRecordHeader;
	int retCode;
	long messageLength;
	queueDetails_t *queueDetailsP;
	bool_t moreLogMessages;

    TRACE_ENTRY("logRemoveEntries", ("aggregateId=%.8x, commit=%s", 
		                              aggregateId, commit ? "TRUE" : "FALSE"));

	retCode = getLogQueueDetails(&queueDetailsP, errorStringP);

	if (retCode != TT_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to get recovery queue details"));
		TRACE_RETURN(RSN_005, TT_FAILED_TO_ACCESS_RECOVERY_MEDIA);
	}

	/* set CorrelId value to the name of the execution group */
	memset(correlId, '\0', MQ_CORREL_ID_LENGTH);
	if (strlen(executionGroupName) > MQ_CORREL_ID_LENGTH)
	{
		/* take the first and last 12 chars of execution group name */
		memcpy(correlId, executionGroupName, 
			   MQ_CORREL_ID_LENGTH/2);
		memcpy(correlId+MQ_CORREL_ID_LENGTH/2, 
			   executionGroupName+strlen(executionGroupName)-MQ_CORREL_ID_LENGTH/2, 
			   MQ_CORREL_ID_LENGTH/2);
	}
	else
	{
		memcpy(correlId, executionGroupName, strlen(executionGroupName));
	}

	/* 
	 * Read message as part of UOW.
	 * When called from an MQSI thread, this will be part of
	 * the message flow transaction 
	 */
	gmo.Options += MQGMO_SYNCPOINT;

	/* don't wait for messages */
	gmo.Options += MQGMO_NO_WAIT;

	/* not interested in the actual message */
	gmo.Options += MQGMO_ACCEPT_TRUNCATED_MSG;

	/* set the msg id to the aggregate id */
	memcpy(md.MsgId, aggregateId, MQ_MSG_ID_LENGTH);
	TRACE_MID(TRACE_INFO, ("Retrieving messages with MsgId value..."));
    TRACE_MIDHEX(TRACE_INFO, md.MsgId, MQ_MSG_ID_LENGTH);

	/* set the correlation id to the execution group name */
	memcpy(md.CorrelId, correlId, MQ_CORREL_ID_LENGTH);
	TRACE_MID(TRACE_INFO, ("Retrieving messages with CorrelId value..."));
    TRACE_MIDHEX(TRACE_INFO, md.CorrelId, MQ_CORREL_ID_LENGTH);

	/* read all messages from the queue which belong to this aggregate */
	moreLogMessages = TRUE;
	while(moreLogMessages)
	{
		MQGET(queueDetailsP->qmConn,        /* connection handle */
			  queueDetailsP->queueHandle,   /* object handle */
			  &md,                          /* message descriptor */
			  &gmo,                         /* get message options */
			  sizeof(logRecordHeader),      /* size of receive buffer */
	  (char *)&logRecordHeader,             /* message buffer */
	  (long *)&messageLength,               /* returned message length */
			  &compCode,                    /* completion code */
			  &reason);                     /* reason code */

		if (reason == MQRC_NO_MSG_AVAILABLE)
		{
			TRACE_MID(TRACE_INFO, ("No more messages available"));
			moreLogMessages = FALSE;
			continue;
		}
 
		if (reason == MQRC_TRUNCATED_MSG_ACCEPTED)
		{
			TRACE_MID(TRACE_INFO, ("Log message read from queue, type is %s",
				                   logRecordHeader.recordType == NEW_INSTANCE ? "NEW_INSTANCE" :
				                   logRecordHeader.recordType == MEMBER ? "MEMBER" :
			                       logRecordHeader.recordType == REPLY ? "REPLY" :
			                       logRecordHeader.recordType == FINISH ? "FINISH" :
                                   "'unknown'"));
			continue;
		}
 
		if (reason != MQRC_NONE)
		{
			TRACE_MID(TRACE_ERROR, ("MQGET for '%s' failed with "
								  "reason code %ld",
								  AGGREGATOR_RECOVERY_QUEUE,
								  reason));
			TRACE_RETURN(RSN_005, TT_FAILED_TO_READ_LOG_RECORD);
		}
	}

	/* commit this work if requested */
	if (commit)
	{
		TRACE_MID(TRACE_INFO, ("Committing messages"));
		MQCMIT(queueDetailsP->qmConn,
		       &compCode,            /* completion code */
		       &reason);             /* reason code */

		/* All non-ok response are significant for this command */
		if (compCode != MQCC_OK)
		{
			/* raise an exception...needs investigation */
		}
	}

    TRACE_RETURN(RSN_001, TT_SUCCESS);
}

typedef char aggregateId_t[MQ_CORREL_ID_LENGTH];
/******************************************************************************/
/* truncateLog                                                                */
/******************************************************************************/
threadFunction_t truncateLog(IN void *dataP)
{

    MQLONG   compCode,                  /* completion code               */
             reason;                    /* reason code                   */
    MQOD od = {MQOD_DEFAULT};           /* Object Descriptor */
    MQMD md,                            /* Message Descriptor */
         mdDefault = {MQMD_DEFAULT};    /* Message Descriptor default values */
    MQGMO gmo,                          /* get message options */
          gmoDefault = {MQGMO_DEFAULT}; /* get message options default values */
	char correlId[MQ_CORREL_ID_LENGTH],
		 *logRecord;
	char *completedAggregateIdP = NULL;
	int numCompletedAggregateIds = 0,
	    numSpareCompletedAggregateIds = 0,
		i,
		messageLength,
		retCode;
	logRecordHeader_t logRecordHeader,
		              *logRecordHeaderP;
	long bufferSize;
	queueDetails_t *queueDetailsP;
	bool_t moreLogMessages,
		   finishRecordFound;
    logTruncationRequest_t *logTruncReqP,
		                   *currentTruncReqP,
		                   *nextTruncReqP;
	int timeToWait;
	char *errMsg;

    TRACE_ENTRY("truncateLog", ("dataP=%.8x", dataP));

	/* get handle to recovery queue */
	retCode = getLogQueueDetails(&queueDetailsP, &errMsg);

	if (retCode != TT_SUCCESS)
	{
		/* create error message to pass back to parent */
        TRACE_MID(TRACE_INFO, ("%s", errMsg));
        logThreadErrorMessage = strdup(errMsg);

		/* lock the alarm thread sync mutex and post the parent thread */
	    threadMutexLock(logThreadSyncMutex);
		threadCondSignal(logThreadSyncCondVar);
	    threadMutexUnlock(logThreadSyncMutex);

        TRACE_RETURN_TYPE(RSN_002, FUNCTION_RETURN_TYPE, 0);
	}

	/* let the parent thread know that we're up and running */
	logThreadInitialised = TRUE;

	/* lock the alarm thread sync mutex and post the parent thread */
    TRACE_MID(TRACE_INFO, ("About to signal parent for initialisation complete"));
	threadMutexLock(logThreadSyncMutex);
	threadCondSignal(logThreadSyncCondVar);
	threadMutexUnlock(logThreadSyncMutex);

	/* Periodically, remove complete aggregates from the recovery queue */

	/* browse the recovery queue looking for log records with 
	 *  - correlation id value of this execution group
	 */

	/* set CorrelId value to the name of the execution group */
	memset(correlId, '\0', MQ_CORREL_ID_LENGTH);
	if (strlen(executionGroupName) > MQ_CORREL_ID_LENGTH)
	{
		/* take the first and last 12 chars of execution group name */
		memcpy(correlId, executionGroupName, 
			   MQ_CORREL_ID_LENGTH/2);
		memcpy(correlId+MQ_CORREL_ID_LENGTH/2, 
			   executionGroupName+strlen(executionGroupName)-MQ_CORREL_ID_LENGTH/2, 
			   MQ_CORREL_ID_LENGTH/2);
	}
	else
	{
		memcpy(correlId, executionGroupName, strlen(executionGroupName));
	}
	/* set the correlation id to the execution group name */
	memcpy(mdDefault.CorrelId, correlId, MQ_CORREL_ID_LENGTH);

	/* don't wait for messages */
	gmoDefault.Options |= MQGMO_NO_WAIT;
	/* not interested in the whole message */
	gmoDefault.Options |= MQGMO_ACCEPT_TRUNCATED_MSG;

	/* allocate the initial buffer size */
	logRecord = (char *)malloc(DEFAULT_LOG_RECORD_SIZE);
	TRACE_ALLOC("log.c-15", logRecord);
	bufferSize = DEFAULT_LOG_RECORD_SIZE;

	/* keep processing recovery queue until requested to shutdown */
	while(!shutdownRequested)
	{
		/* go to sleep for 1 minute or until we're signalled, whichever comes first */
		timeToWait = LOG_TRUNCATION_INTERVAL; 
        TRACE_MID(TRACE_INFO, ("Next expiry interval time is %dms", timeToWait));

#if 0
	    /* wait for the timer to pop or someone to wake us */
		threadMutexLock(logThreadWorkMutex);

        TRACE_MID(TRACE_INFO, ("Log thread about to sleep for expiry interval"));
		retCode = threadCondTimedWait(logThreadWorkCondVar, timeToWait);
        TRACE_MID(TRACE_INFO, ("Log thread woken, retCode=%s",
			retCode==THREAD_OK ? "THREAD_OK" :
			retCode==THREAD_TIMED_OUT ? "THREAD_TIMED_OUT" : "unknown rc"));
		threadMutexUnlock(logThreadWorkMutex);
#else
		/* 
		 * The work of the log thread is currently being done inline.  This thread is
		 * now superfluous, so we'll just keep it dormant until its time to shutdown
	     */
		threadMutexLock(logThreadWorkMutex);

        TRACE_MID(TRACE_INFO, ("Log thread about to sleep for expiry interval"));
		retCode = threadCondWait(logThreadWorkCondVar);
        TRACE_MID(TRACE_INFO, ("Log thread woken, retCode=%s",
			retCode==THREAD_OK ? "THREAD_OK" :
			retCode==THREAD_TIMED_OUT ? "THREAD_TIMED_OUT" : "unknown rc"));
		threadMutexUnlock(logThreadWorkMutex);
#endif
		if (shutdownRequested)
		{
			continue;
		}

		/* 
		 * If we've been signalled, then process the truncation request list, otherwise,
		 * just do a browse through the log checking for any FINISH records that might 
		 * still exist.
		 */
		if (retCode == THREAD_OK)
		{
			/* check if there's any truncation requests outstanding */
			threadMutexLock(logThreadWorkMutex);
			if (numLogTruncationRequests)
			{
				logTruncReqP = logTruncationRequestAnchor.nextP;
				logTruncationRequestAnchor.nextP = NULL;
				numLogTruncationRequests = 0;
			}
			threadMutexUnlock(logThreadWorkMutex);

			/* 
			 * Run down the request list and for each entry, remove all log records with
			 * matching aggregate id.  However, only commit iff a FINISH record exists.
			 */
			currentTruncReqP = logTruncReqP;
			while(currentTruncReqP)
			{
				nextTruncReqP = currentTruncReqP->nextP;

				/* 
				 * Read message as part of UOW.
				 * When called from an MQSI thread, this will be part of
				 * the message flow transaction 
				 */
				memcpy(&gmo, &gmoDefault, sizeof(gmo));
				memcpy(&md, &mdDefault, sizeof(md));

				gmo.Options |= MQGMO_SYNCPOINT;

				/* set the msg id to the aggregate id */
				memcpy(md.MsgId, currentTruncReqP->aggregateId, MQ_MSG_ID_LENGTH);
				TRACE_MID(TRACE_INFO, ("Retrieving messages with MsgId value..."));
				TRACE_MIDHEX(TRACE_INFO, md.MsgId, MQ_MSG_ID_LENGTH);

				TRACE_MID(TRACE_INFO, ("Retrieving messages with CorrelId value..."));
				TRACE_MIDHEX(TRACE_INFO, md.CorrelId, MQ_CORREL_ID_LENGTH);

				/* read all messages from the queue which belong to this aggregate */
				moreLogMessages = TRUE;
				finishRecordFound = FALSE;
				while(moreLogMessages)
				{
					MQGET(queueDetailsP->qmConn,        /* connection handle */
						  queueDetailsP->queueHandle,   /* object handle */
						  &md,                          /* message descriptor */
						  &gmo,                         /* get message options */
						  sizeof(logRecordHeader),      /* size of receive buffer */
				  (char *)&logRecordHeader,             /* message buffer */
				  (long *)&messageLength,               /* returned message length */
						  &compCode,                    /* completion code */
						  &reason);                     /* reason code */

					if (reason == MQRC_NO_MSG_AVAILABLE)
					{
						TRACE_MID(TRACE_INFO, ("No more messages available"));
						moreLogMessages = FALSE;
						continue;
					}
 
					if (reason == MQRC_TRUNCATED_MSG_ACCEPTED ||
						reason == MQRC_NONE)
					{
						TRACE_MID(TRACE_INFO, ("Log message read from queue, type is %s",
											   logRecordHeader.recordType == NEW_INSTANCE ? "NEW_INSTANCE" :
											   logRecordHeader.recordType == MEMBER ? "MEMBER" :
											   logRecordHeader.recordType == REPLY ? "REPLY" :
											   logRecordHeader.recordType == FINISH ? "FINISH" :
											   "'unknown'"));
						if (logRecordHeader.recordType == FINISH)
						{
							finishRecordFound = TRUE;
							break;
						}
						continue;
					}
 
					if (reason != MQRC_NONE)
					{
						TRACE_MID(TRACE_ERROR, ("MQGET for '%s' failed with "
											  "reason code %ld",
											  AGGREGATOR_RECOVERY_QUEUE,
											  reason));
				        TRACE_RETURN_TYPE(RSN_002, FUNCTION_RETURN_TYPE, 0);
					}
				}

				/* only commit the reads if a FINISH record was present */
				if (finishRecordFound)
				{
					TRACE_MID(TRACE_INFO, ("Committing truncated log messages"));
					MQCMIT(queueDetailsP->qmConn,
						   &compCode,            /* completion code */
						   &reason);             /* reason code */

					/* All non-ok response are significant for this command */
					if (compCode != MQCC_OK)
					{
						/* raise an exception...needs investigation */
					}
				}
				else
				{
					TRACE_MID(TRACE_INFO, ("Backing out truncated log messages"));
					MQBACK(queueDetailsP->qmConn,
						   &compCode,            /* completion code */
						   &reason);             /* reason code */

					/* All non-ok response are significant for this command */
					if (compCode != MQCC_OK)
					{
						/* raise an exception...needs investigation */
					}

					/* place request back in list for next time */
					threadMutexLock(logThreadWorkMutex);
					INSERT_AFTER_IN_LINKED_LIST(logTruncationRequestAnchorP, currentTruncReqP);
					numLogTruncationRequests++;
					threadMutexUnlock(logThreadWorkMutex);
				}

				/* move on to the next truncation request */
				currentTruncReqP = nextTruncReqP;
			}
		}
		else /* normal 'timed-out processing */
		{
			/* reset the completed aggregates table */
			numSpareCompletedAggregateIds += numCompletedAggregateIds;
			numCompletedAggregateIds = 0;

			/* browse through the queue looking for any aggregate complete records */
			TRACE_MID(TRACE_INFO, ("About to browse recovery queue for aggregate "
								   "FINISH records"));

			memcpy(&gmo, &gmoDefault, sizeof(gmo));

			/* start at first message in the recovery queue */
			gmo.Options |= MQGMO_BROWSE_FIRST;

			memcpy(&md, &mdDefault, sizeof(md));
			TRACE_MID(TRACE_INFO, ("Retrieving messages with CorrelId value..."));
			TRACE_MIDHEX(TRACE_INFO, md.CorrelId, MQ_CORREL_ID_LENGTH);

			/* read all messages from the queue which belong to this execution group */
			moreLogMessages = TRUE;
			while(moreLogMessages)
			{
				MQGET(queueDetailsP->qmConn,        /* connection handle */
					  queueDetailsP->queueHandle,   /* object handle */
					  &md,                          /* message descriptor */
					  &gmo,                         /* get message options */
					  bufferSize,                   /* size of receive buffer */
					  logRecord,                    /* receive buffer */
			  (long *)&messageLength,               /* returned message length */
					  &compCode,                    /* completion code */
					  &reason);                     /* reason code */

				if (reason == MQRC_NO_MSG_AVAILABLE)
				{
					TRACE_MID(TRACE_INFO, ("No more messages available"));
					moreLogMessages = FALSE;
					continue;
				}
 
				/* any other error is assumed to be a serious one */
				if (reason != MQRC_NONE && reason != MQRC_TRUNCATED_MSG_ACCEPTED)
				{
					TRACE_MID(TRACE_ERROR, ("MQGET for '%s' failed with "
										  "reason code %ld",
										  AGGREGATOR_RECOVERY_QUEUE,
										  reason));
					if (completedAggregateIdP)
					{
						free(completedAggregateIdP);
						TRACE_FREE("log.c-27", completedAggregateIdP);
					}
					free(logRecord);
					TRACE_FREE("log.c-26", logRecord);
					TRACE_RETURN_TYPE(RSN_003, FUNCTION_RETURN_TYPE, 0);
				}

				/* make sure we get the next record next time round */
				memcpy(&gmo, &gmoDefault, sizeof(gmo));
				gmo.Options |= MQGMO_BROWSE_NEXT;

				/* we're only interested in aggregate FINISH records at this stage */
				logRecordHeaderP = (logRecordHeader_t *)logRecord;
		
				/* process the log record just read */
				TRACE_MID(TRACE_INFO, ("Log message read from queue, type is %s",
									   logRecordHeaderP->recordType == NEW_INSTANCE ? "NEW_INSTANCE" :
									   logRecordHeaderP->recordType == MEMBER ? "MEMBER" :
									   logRecordHeaderP->recordType == REPLY ? "REPLY" :
									   logRecordHeaderP->recordType == FINISH ? "FINISH" :
									   "'unknown'"));

				/* make a note of the aggregate id if this is an aggregate FINISH record */
				if (logRecordHeaderP->recordType != FINISH)
				{
					continue;
				}

				/* increase number of slots if there's none spare */
				if (!numSpareCompletedAggregateIds)
				{
					if (completedAggregateIdP)
					{
						completedAggregateIdP = (char *)realloc(completedAggregateIdP, 
									   sizeof(aggregateId_t) * 
											(numSpareCompletedAggregateIds + 10));
						TRACE_ALLOC("log.c-16", completedAggregateIdP);
					}
					else
					{
						completedAggregateIdP = (char *)malloc(sizeof(aggregateId_t) * 10);
						TRACE_ALLOC("log.c-16", completedAggregateIdP);
					}
					numSpareCompletedAggregateIds += 10;
				}

				/* add this aggregate id to our list of completed aggregates */
				memcpy(completedAggregateIdP+(numCompletedAggregateIds*sizeof(aggregateId_t)), 
					   logRecordHeaderP->aggregateId, MQ_CORREL_ID_LENGTH);
				numCompletedAggregateIds++;
				numSpareCompletedAggregateIds--;
				continue;
			}

			/* 
			 * Run down the list of completed aggregates and for each aggregate, remove all
			 * log records that belong to it.
			 */
			TRACE_MID(TRACE_INFO, ("Found %d aggregate FINISH record%s",
									numCompletedAggregateIds,
									numCompletedAggregateIds==1 ? "" : "s"));
			for (i=0; i<numCompletedAggregateIds; i++)
			{
				/* read all records which match this aggregate id and commit */
				retCode = logRemoveEntries(completedAggregateIdP+i, TRUE, &errMsg);

				if (retCode != TT_SUCCESS)
				{
					/* generate a log message, not fatal */

				}
			}
		}
	}

	TRACE_MID(TRACE_INFO, ("About to issue MQDISC for '%s'",
							queueManagerName));
	MQDISC(&queueDetailsP->qmConn,                 /* connection handle */
		   &compCode,               /* completion code */
		   &reason);                /* reason code */

	/* free up the log record buffer */
	free(logRecord);
    TRACE_FREE("log.c-26", logRecord);

    TRACE_MID(TRACE_INFO, ("Log thread terminating"));

    TRACE_RETURN_TYPE(RSN_003, FUNCTION_RETURN_TYPE, 0);
}


/*
 * internal functions 
 */

/******************************************************************************/
/* writeLog                                                                   */
/******************************************************************************/
int writeLog(IN char *logMessage,
			 IN int messageLength,
			 IN char *aggregateId,
	 		 OUT char **errorStringP)
{

    MQLONG   compCode,                  /* completion code               */
             reason;                    /* reason code                   */
    MQOD od = {MQOD_DEFAULT};           /* Object Descriptor */
    MQMD md = {MQMD_DEFAULT};           /* Message Descriptor */
    MQPMO pmo = {MQPMO_DEFAULT};        /* put message options */
	int retCode;
	queueDetails_t *queueDetailsP;

    TRACE_ENTRY("writeLog", ("logMessage=%.8x, messageLength=%d",
		                      logMessage, messageLength));

	retCode = getLogQueueDetails(&queueDetailsP, errorStringP);

	if (retCode != TT_SUCCESS)
	{
		TRACE_MID(TRACE_ERROR, ("Failed to get recovery queue details"));
		TRACE_RETURN(RSN_005, TT_FAILED_TO_ACCESS_RECOVERY_MEDIA);
	}
	
	/* set CorrelId value to the name of the execution group */
	if (strlen(executionGroupName) > MQ_CORREL_ID_LENGTH)
	{
		/* take the first and last 12 chars of execution group name */
		memcpy(md.CorrelId, executionGroupName, 
			   MQ_CORREL_ID_LENGTH/2);
		memcpy(md.CorrelId+MQ_CORREL_ID_LENGTH/2, 
			   executionGroupName+strlen(executionGroupName)-MQ_CORREL_ID_LENGTH/2, 
			   MQ_CORREL_ID_LENGTH/2);
	}
	else
	{
		memcpy(md.CorrelId, executionGroupName, strlen(executionGroupName));
	}

	/* commit message as part of message flow transaction */
	pmo.Options += MQPMO_SYNCPOINT;

	/* write persistent log message */
    md.Persistence = MQPER_PERSISTENT;

	/* set the msg id to the aggregate id */
	memcpy(md.MsgId, aggregateId, MQ_MSG_ID_LENGTH);

    TRACE_MID(TRACE_INFO, ("About to write log message"));
    
    /* put the message on the queue */
    MQPUT(queueDetailsP->qmConn,    /* connection handle */
          queueDetailsP->queueHandle, /* object handle */
          &md,                      /* message descriptor */
          &pmo,                     /* put options */
          messageLength,            /* zero msg length */
          logMessage,               /* log message buffer */
          &compCode,                /* completion code */
          &reason);                 /* reason code */
        
	if (compCode != MQCC_OK)
	{
		TRACE_MID(TRACE_ERROR, ("MQPUT for '%s' failed with "
							  "reason code %ld",
							  AGGREGATOR_RECOVERY_QUEUE,
							  reason));
        TRACE_RETURN(RSN_005, TT_FAILED_TO_WRITE_LOG_RECORD);
	}
    TRACE_MID(TRACE_INFO, ("Log message written to queue"));

    TRACE_RETURN(RSN_001, TT_SUCCESS);
}

/******************************************************************************/
/* getLogQueueDetails                                                         */
/******************************************************************************/
int getLogQueueDetails(IN queueDetails_t **queueDetailsPP,
					   OUT char **errorMessageP)
{

    MQLONG   compCode,                  /* completion code               */
             reason;                    /* reason code                   */
    MQLONG   openOptions;               /* MQOPEN options                */
    MQOD od = {MQOD_DEFAULT};           /* Object Descriptor */
    MQMD md = {MQMD_DEFAULT};           /* Message Descriptor */
    MQPMO pmo = {MQPMO_DEFAULT};        /* put message options */
	char *queueNameP;
	int retCode;
	char errorBuffer[1000];
	queueDetails_t *queueDetailsP;

    TRACE_ENTRY("getLogQueueDetails", ("queueDetailsPP=%.8x", 
		                             queueDetailsPP));

	/* check if thread has already opened the logging queue */
	retCode = threadGetSpecificStorage(loggingKey,
		                               (void **)&queueDetailsP);
    
	if (!queueDetailsP)
	{
		TRACE_MID(TRACE_INFO, ("Thread has no handle to recovery queue"));
        queueDetailsP = (queueDetails_t *)malloc(sizeof(queueDetails_t));
		TRACE_ALLOC("log.c-17", queueDetailsP);
		TRACE_MID(TRACE_INFO, ("Thread specific storage allocated at %8x", 
			                   queueDetailsP));

		/* queue not opened yet... */

		/* get handle to queue manager - should return ALREADY_CONNECTED */
		TRACE_MID(TRACE_INFO, ("About to issue MQCONN for '%s'",
								queueManagerName));
		MQCONN(queueManagerName,
			   &queueDetailsP->qmConn,   /* connection handle */
			   &compCode,               /* completion code */
			   &reason);                /* reason code */

		if (compCode == MQCC_FAILED)
		{
			/* create error message to pass back to parent */
			sprintf(errorBuffer, "MQCONN for queue manager '%s' "
							    "failed with reason code %ld",
							    queueManagerName, reason);
			TRACE_MID(TRACE_ERROR, ("%s", errorBuffer));
			if (errorMessageP)
			{
				*errorMessageP = (char *)malloc(strlen(errorBuffer)+1);
				TRACE_ALLOC("log.c-18", errorMessageP);
				strcpy(*errorMessageP, errorBuffer);
			}

			TRACE_RETURN(RSN_001, TT_FAILED_TO_CONNECT_TO_QUEUE_MANAGER);
		}

		/* open the recovery queue */
		queueNameP = AGGREGATOR_RECOVERY_QUEUE;

		strncpy(od.ObjectName, 
                queueNameP,
				(size_t)MQ_Q_NAME_LENGTH);

        openOptions = 0;
		openOptions += MQOO_OUTPUT;        /* open for writing */
		openOptions += MQOO_INPUT_SHARED;  /* open for reading */
		openOptions += MQOO_BROWSE;        /* open for browsing */
		openOptions += MQOO_FAIL_IF_QUIESCING;     /* but not if MQM stopping */
		 
		TRACE_MID(TRACE_INFO, ("About to issue MQOPEN for '%s'",
								queueNameP));

		MQOPEN(queueDetailsP->qmConn, /* connection handle */
			   &od,                 /* object descriptor for queue */
			   openOptions,         /* open options */
			   &queueDetailsP->queueHandle, /* object handle */
			   &compCode,           /* completion code */
			   &reason);            /* reason code */
            
		if (compCode == MQCC_FAILED)
		{
			/* create error message to pass back to parent */
			sprintf(errorBuffer, "MQOPEN for '%s' failed with "
							  "reason code %ld", queueNameP, reason);
			TRACE_MID(TRACE_ERROR, ("%s", errorBuffer));
			if (errorMessageP)
			{
				*errorMessageP = (char *)malloc(strlen(errorBuffer)+1);
				TRACE_ALLOC("log.c-19", errorMessageP);
				strcpy(*errorMessageP, errorBuffer);
			}
            TRACE_RETURN(RSN_004, TT_FAILED_TO_OPEN_RECOVERY_QUEUE);
		}

		/* store queue details in thread slot */
		retCode = threadSetSpecificStorage(loggingKey,
										   (void *)queueDetailsP);
	}

	*queueDetailsPP = queueDetailsP;
    TRACE_RETURN(RSN_001, TT_SUCCESS);
}

